import '@/autogen/api-client-jsdoc'

import type { Endpoints, SwitchCaseResponseType } from '@/api.types'

export type ApiError = {
  id: string
  code: string
  info: Record<string, any>
  kind: 'client' | 'server'
  message: string
}

export type FetchLike = (
  input: string,
  init?: {
    body?: string
    cache?: RequestCache
    credentials?: RequestCredentials
    headers?: Record<string, string>
    method?: string
  }
) => Promise<{
  json(): Promise<any>
  status: number
}>

export class ApiClient {
  public credential: string | null | undefined
  public fetch: FetchLike
  public origin: string

  constructor(opts: {
    credential?: ApiClient['credential']
    fetch?: ApiClient['fetch'] | null
    origin: ApiClient['origin']
  }) {
    this.credential = opts.credential
    this.fetch = opts.fetch ?? ((...args) => fetch(...args))
    this.origin = opts.origin
  }

  public request<E extends keyof Endpoints, P extends Endpoints[E]['req']>(
    endpoint: E,
    params: P = {} as P,
    credential?: string | null
  ) {
    return new Promise<SwitchCaseResponseType<E, P>>((resolve, reject) => {
      this.fetch(`${this.origin}/api/${endpoint}`, {
        body: JSON.stringify({
          ...params,
          i: credential !== undefined ? credential : this.credential
        }),
        cache: 'no-cache',
        credentials: 'omit',
        headers: {
          'Content-Type': 'application/json'
        },
        method: 'POST'
      })
        .then(async (res) => {
          const body = res.status === 204 ? null : await res.json()

          if (res.status === 200 || res.status === 204) {
            resolve(body)
          } else {
            reject({
              [AKKO_API_ERROR]: true,
              ...body.error
            })
          }
        })
        .catch(reject)
    })
  }
}

const AKKO_API_ERROR = Symbol()

export function isApiError(reason: any): reason is ApiError {
  return reason[AKKO_API_ERROR] === true
}

export type { SwitchCaseResponseType }
