import SwaggerParser from '@apidevtools/swagger-parser'
import { mkdir, writeFile } from 'node:fs/promises'
import openapiTS from 'openapi-typescript'
import { toPascal } from 'ts-case-convert'
import { OpenAPIV3 } from 'openapi-types'

enum OperationsAliasType {
  REQUEST = 'Request',
  RESPONSE = 'Response'
}

interface IOperationTypeAlias {
  generateName(): string
  toLine(): string
  readonly type: OperationsAliasType
}

class EmptyTypeAlias implements IOperationTypeAlias {
  public readonly type: OperationsAliasType

  constructor(type: OperationsAliasType) {
    this.type = type
  }

  public generateName() {
    return `Empty${this.type}`
  }

  public toLine() {
    const name = this.generateName()

    return `export type ${name} = Record<string, unknown> | undefined`
  }
}

class Endpoint {
  public readonly operationId: string
  public request?: IOperationTypeAlias
  public response?: IOperationTypeAlias

  constructor(operationId: string) {
    this.operationId = operationId
  }

  public toLine() {
    const reqName = this.request?.generateName() ?? emptyRequest.generateName()
    const resName = this.response?.generateName()
      ?? emptyResponse.generateName()

    return `'${this.operationId}': { req: ${reqName}; res: ${resName} }`
  }
}

class OperationTypeAlias implements IOperationTypeAlias {
  public readonly mediaType: string
  public readonly operationId: string
  public readonly type: OperationsAliasType

  constructor(
    operationId: string,
    mediaType: string,
    type: OperationsAliasType
  ) {
    this.mediaType = mediaType
    this.operationId = operationId
    this.type = type
  }

  public generateName() {
    const nameBase = this.operationId.replace(/\//g, '-')

    return toPascal(nameBase + this.type)
  }

  public toLine() {
    const name = this.generateName()

    return this.type === OperationsAliasType.REQUEST
      ? `export type ${name} = operations['${this.operationId}']['requestBody']['content']['${this.mediaType}']`
      : `export type ${name} = operations['${this.operationId}']['responses']['200']['content']['${this.mediaType}']`
  }
}

const emptyRequest = new EmptyTypeAlias(OperationsAliasType.REQUEST)
const emptyResponse = new EmptyTypeAlias(OperationsAliasType.RESPONSE)

function filterUndefined<T>(item: T): item is Exclude<T, undefined> {
  return item !== undefined
}

async function generateApiClientJsDoc(
  openApiDocs: OpenAPIV3.Document,
  apiClientFileName: string,
  endpointsFileName: string,
  warningsOutputPath: string
) {
  const endpoints = [] as { description: string; operationId: string }[]

  const paths = openApiDocs.paths
  const postPathItems = Object.keys(paths)
    .map(it => paths[it]?.post)
    .filter(filterUndefined)

  for (const operation of postPathItems) {
    const operationId = operation.operationId!

    if (operation.description) {
      endpoints.push({
        description: operation.description,
        operationId: operationId
      })
    }
  }

  const endpointOutputLine = [] as string[]

  endpointOutputLine.push(generateVersionHeaderComment(openApiDocs))
  endpointOutputLine.push('')

  endpointOutputLine.push(
    `import type { SwitchCaseResponseType } from '${toImportPath(apiClientFileName)}'`
  )

  endpointOutputLine.push(
    `import type { Endpoints } from '${toImportPath(endpointsFileName)}'`
  )

  endpointOutputLine.push('')

  endpointOutputLine.push(
    `declare module '${toImportPath(apiClientFileName)}' {`
  )

  endpointOutputLine.push('  export interface ApiClient {')

  for (let i = 0; i < endpoints.length; i++) {
    const endpoint = endpoints[i]

    endpointOutputLine.push(
      '    /**',
      `     * ${endpoint.description.split('\n').join('\n     * ')}`,
      '     */',
      `    request<E extends '${endpoint.operationId}', P extends Endpoints[E]['req']>(`,
      '      endpoint: E,',
      '      params: P,',
      '      credential?: string | null',
      '    ): Promise<SwitchCaseResponseType<E, P>>'
    )

    if (i < endpoints.length - 1) {
      endpointOutputLine.push('\n')
    }
  }

  endpointOutputLine.push('  }')
  endpointOutputLine.push('}')
  endpointOutputLine.push('')

  await writeFile(warningsOutputPath, endpointOutputLine.join('\n'))
}

async function generateBaseTypes(
  openApiDocs: OpenAPIV3.Document,
  openApiJsonPath: string,
  typeFileName: string
) {
  const disabledLints = [
    '@typescript-eslint/naming-convention',
    '@typescript-eslint/no-explicit-any'
  ]

  const lines = [] as string[]

  for (const lint of disabledLints) {
    lines.push(`/* eslint ${lint}: 0 */`)
  }

  lines.push('')
  lines.push(generateVersionHeaderComment(openApiDocs))
  lines.push('')

  const generatedTypes = await openapiTS(openApiJsonPath, { exportType: true })

  lines.push(generatedTypes)
  lines.push('')

  await writeFile(typeFileName, lines.join('\n'))
}

async function generateEndpoints(
  openApiDocs: OpenAPIV3.Document,
  typeFileName: string,
  entitiesOutputPath: string,
  endpointOutputPath: string
) {
  const endpoints = [] as Endpoint[]

  const paths = openApiDocs.paths
  const postPathItems = Object.keys(paths)
    .map(it => paths[it]?.post)
    .filter(filterUndefined)

  for (const operation of postPathItems) {
    const operationId = operation.operationId!

    const endpoint = new Endpoint(operationId)

    endpoints.push(endpoint)

    if (isRequestBodyObject(operation.requestBody)) {
      const reqContent = operation.requestBody.content
      const supportMediaTypes = Object.keys(reqContent)

      if (supportMediaTypes.length > 0) {
        endpoint.request = new OperationTypeAlias(
          operationId,
          supportMediaTypes[0],
          OperationsAliasType.REQUEST
        )
      }
    }

    if (
      isResponseObject(operation.responses['200'])
        && operation.responses['200'].content
    ) {
      const resContent = operation.responses['200'].content
      const supportMediaTypes = Object.keys(resContent)

      if (supportMediaTypes.length > 0) {
        endpoint.response = new OperationTypeAlias(
          operationId,
          supportMediaTypes[0],
          OperationsAliasType.RESPONSE
        )
      }
    }
  }

  const entitiesOutputLine = [] as string[]

  entitiesOutputLine.push(generateVersionHeaderComment(openApiDocs))
  entitiesOutputLine.push('')

  entitiesOutputLine.push(
    `import { operations } from '${toImportPath(typeFileName)}'`
  )

  entitiesOutputLine.push('')

  entitiesOutputLine.push(
    new EmptyTypeAlias(OperationsAliasType.REQUEST).toLine()
  )

  entitiesOutputLine.push(
    new EmptyTypeAlias(OperationsAliasType.RESPONSE).toLine()
  )

  entitiesOutputLine.push('')

  const entities = endpoints
    .flatMap(it => [it.request, it.response].filter(i => i))
    .filter(filterUndefined)

  entitiesOutputLine.push(...entities.map(it => it.toLine()))
  entitiesOutputLine.push('')

  await writeFile(entitiesOutputPath, entitiesOutputLine.join('\n'))

  const endpointOutputLine = [] as string[]

  endpointOutputLine.push(generateVersionHeaderComment(openApiDocs))
  endpointOutputLine.push('')

  endpointOutputLine.push('import type {')

  endpointOutputLine.push(
    ...[
      emptyRequest,
      emptyResponse,
      ...entities
    ].map(it => `  ${it.generateName()},`)
  )

  endpointOutputLine.push(`} from '${toImportPath(entitiesOutputPath)}'`)
  endpointOutputLine.push('')

  endpointOutputLine.push('export type Endpoints = {')

  endpointOutputLine.push(
    ...endpoints.map(it => `  ${it.toLine()}`)
  )

  endpointOutputLine.push('}')
  endpointOutputLine.push('')

  await writeFile(endpointOutputPath, endpointOutputLine.join('\n'))
}

async function generateSchemaEntities(
  openApiDocs: OpenAPIV3.Document,
  typeFileName: string,
  outputPath: string
) {
  if (!openApiDocs.components?.schemas) {
    return
  }

  const schemas = openApiDocs.components.schemas

  const schemaNames = Object.keys(schemas)
  const typeAliasLines = [] as string[]

  typeAliasLines.push(generateVersionHeaderComment(openApiDocs))
  typeAliasLines.push('')

  typeAliasLines.push(
    `import { components } from '${toImportPath(typeFileName)}'`
  )

  typeAliasLines.push('')

  typeAliasLines.push(
    ...schemaNames.map(
      it => `export type ${it} = components['schemas']['${it}']`
    )
  )

  typeAliasLines.push('')

  await writeFile(outputPath, typeAliasLines.join('\n'))
}

function generateVersionHeaderComment(openApiDocs: OpenAPIV3.Document) {
  const contents = {
    'Generated at': new Date().toISOString(),
    Version: openApiDocs.info.version
  }

  const lines = [] as string[]

  lines.push('//')

  for (const [key, value] of Object.entries(contents)) {
    lines.push(`// ${key}: ${value}`)
  }

  lines.push('//')

  return lines.join('\n')
}

function isRequestBodyObject(
  value: unknown
): value is OpenAPIV3.RequestBodyObject {
  if (!value) {
    return false
  }

  const {
    content
  } = value as Record<keyof OpenAPIV3.RequestBodyObject, unknown>

  return content !== undefined
}

function isResponseObject(value: unknown): value is OpenAPIV3.ResponseObject {
  if (!value) {
    return false
  }

  const {
    description
  } = value as Record<keyof OpenAPIV3.ResponseObject, unknown>

  return description !== undefined
}

async function main() {
  const generatePath = './built/autogen'

  await mkdir(generatePath, { recursive: true })

  const openApiJsonPath = './api.json'
  const openApiParsed =
    await SwaggerParser.validate(openApiJsonPath) as OpenAPIV3.Document

  const typeFileName = `${generatePath}/types.ts`

  await generateBaseTypes(openApiParsed, openApiJsonPath, typeFileName)

  const modelFileName = `${generatePath}/models.ts`

  await generateSchemaEntities(openApiParsed, typeFileName, modelFileName)

  const endpointFileName = `${generatePath}/endpoint.ts`
  const entitiesFileName = `${generatePath}/entities.ts`

  await generateEndpoints(
    openApiParsed,
    typeFileName,
    entitiesFileName,
    endpointFileName
  )

  const apiClientWarningFileName = `${generatePath}/api-client-jsdoc.ts`

  await generateApiClientJsDoc(
    openApiParsed,
    '../api.ts',
    endpointFileName,
    apiClientWarningFileName
  )
}

function toImportPath(
  fileName: string,
  fromPath = '/built/autogen',
  toPath = ''
) {
  return fileName.replace(fromPath, toPath).replace('.ts', '')
}

main()
