import { getUrlWithLoginId } from '@/utils/get-url-with-login-id'
import { getAccountFromId } from '@/utils/get-account-from-id'
import type { SwMessage, SwMessageOrderType } from '@/types'
import * as Akkordian from 'akkordian.js'

const readBlockingStatus = new Map<string, boolean>()

export const cli = new Akkordian.Api.ApiClient({
  fetch: (...args) => fetch(...args),
  origin
})

export async function api<
  E extends keyof Akkordian.Endpoints,
  O extends Akkordian.Endpoints[E]['req']
>(
  endpoint: E,
  userId?: string,
  options?: O
  // @ts-expect-error
): Promise<ReturnType<typeof cli.request<E, O>> | void> {
  let account: { id: string; token: string } | void = undefined

  if (userId) {
    account = await getAccountFromId(userId)

    if (!account) {
      return
    }
  }

  // @ts-expect-error
  return cli.request(endpoint, options, account?.token)
}

export async function findClient() {
  const clients = await globalThis.clients.matchAll({
    type: 'window'
  })

  return clients.find((c) => !new URL(c.url).searchParams.has('zen')) ?? null
}

export function openAntenna(antennaId: string, loginId: string) {
  return openClient('push', `/timeline/antenna/${antennaId}`, loginId, {
    antennaId
  })
}

export async function openClient(
  order: SwMessageOrderType,
  url: string,
  loginId?: string,
  query: Record<string, SwMessage[string]> = {}
) {
  const client = await findClient()

  if (client) {
    client.postMessage({
      ...query,
      loginId,
      order,
      type: 'order',
      url
    } satisfies SwMessage)

    return client
  }

  return globalThis.clients.openWindow(
    loginId ? getUrlWithLoginId(url, loginId) : url
  )
}

export function openNote(noteId: string, loginId?: string) {
  return openClient('push', `/notes/${noteId}`, loginId, { noteId })
}

export async function openPost(
  options: {
    initialText?: string
    renote?: Akkordian.Entities.Note
    reply?: Akkordian.Entities.Note
  },
  loginId?: string
) {
  const query = new URLSearchParams()
  const url = '/share'

  if (options.initialText) {
    query.set('text', options.initialText)
  }
  if (options.renote) {
    query.set('renoteId', options.renote.id)
  }
  if (options.reply) {
    query.set('replyId', options.reply.id)
  }

  return openClient('post', `${url}?${query}`, loginId, { options })
}

export function openUser(acct: string, loginId?: string) {
  return openClient('push', `/@${acct}`, loginId, { acct })
}

export function sendMarkAllAsRead(userId: string) {
  if (readBlockingStatus.get(userId)) {
    return Promise.resolve()
  }

  readBlockingStatus.set(userId, true)

  return new Promise((res) => {
    setTimeout(() => {
      readBlockingStatus.set(userId, false)

      api('notifications/mark-all-as-read', userId).then(res, res)
    }, 1000)
  })
}
